/*
1.45 sec on Core2 3.16GHz (VC++ 2008, VISTA SP1)
based on "ao.c" @ AOBench
http://lucille.atso-net.jp/aobench/

modifyed by Kuroda Dycoon
(use 4 thread)
1.33 sec on Athlon X2 Dual 4800+ 2.42GHz
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <math.h>

//
#include <windows.h>

#include <vector>

using namespace std;

	class TCriticalS
	{
		public:

#ifdef __GNUC__
			pthread_mutex_t	m;
#else
			CRITICAL_SECTION cs;
#endif

			TCriticalS()
			{
#ifdef __GNUC__
				pthread_mutex_init(&m,NULL);
#else
				InitializeCriticalSection(&cs);
#endif
			}

			~TCriticalS()
			{
				
#ifdef __GNUC__
				pthread_mutex_destroy(&m);
#else
				DeleteCriticalSection(&cs);
#endif
			}

			void Lock()
			{
#ifdef __GNUC__
				pthread_mutex_lock(&m);
#else
				EnterCriticalSection(&cs);
#endif
			}

			void UnLock()
			{
#ifdef __GNUC__
				pthread_mutex_unlock(&m);
#else
				LeaveCriticalSection(&cs);
#endif
			}

//			int TryLock()
//			{
//#ifdef __GNUC__
//				if(pthread_mutex_trylock(&m) != EBUSY)
//				{
//					return True;
//				}
//				else
//				{
//					return 0;
//				}
//#else
//				if(TryEnterCriticalSection(&cs))
//				{
//					return True;
//				}
//				else
//				{
//					return 0;
//				}
//#endif
//			}

#ifdef __GNUC__
			pthread_mutex_t	& GetMutex()
			{
				return m;
			}
#endif
	};

	typedef TCriticalS * PCriticalS;

	inline PCriticalS GetGCriticalSection()
	{
		static TCriticalS cs;
		return &cs;
	}

	class TCriticalSLocker
	{
			PCriticalS cs;
		public:
			TCriticalSLocker(PCriticalS _cs)
			{
				cs = _cs;
				cs->Lock();
			}

			~TCriticalSLocker()
			{
				cs->UnLock();
			}
	
	};

	class TTLSint
	{
		public:
			int TLS;

			inline TTLSint()
			{
				TLS = TlsAlloc();
			}

			inline ~TTLSint()
			{
				TlsFree(TLS);
			}

			inline void Set(int val)
			{
				TlsSetValue(TLS, (void *)(size_t)val);
			}

			inline int Get()
			{
				return (int)(size_t)TlsGetValue(TLS);
			}

			inline int Get() const
			{
				return (int)(size_t)TlsGetValue(TLS);
			}

			inline operator int() const{ return Get(); }

			inline TTLSint& operator =(const int & val)
			{
				Set(val);
				return *this;
			}
	};

	typedef void OnThreadCreateEvent(int idx);
	typedef OnThreadCreateEvent * POnThreadCreateEvent;

	template<class T>class TSingleton
	{
		public:
			
			T * Get()
			{
				static T * obj = NULL;

				if(obj) return obj;
				TCriticalSLocker Locker(GetGCriticalSection());

				if(obj == NULL)
				{
					static T obj2;
					obj = &obj2;
				}
				
				return obj;
			}
	};

	class TOnThreadCreateEvent
	{
		public:
			vector<POnThreadCreateEvent> OnThreadCreate;
			vector<int> ThreadCreated;

			TOnThreadCreateEvent();

			void AddEvent(POnThreadCreateEvent Event);
	};

	extern TSingleton<TOnThreadCreateEvent> GOnThreadCreateEvent;

	void AddOnThreadCreateEvent(POnThreadCreateEvent Event);
	
	/*!
		݂Threadindex߂
	*/
	int GetCurrentThreadIdx();

	class TOnThreadCreate
	{
		public:
			TOnThreadCreate();
			virtual ~TOnThreadCreate();
	};
	
	class TThread;
	typedef TThread * PThread;

	class TVThreadFunc
	{
		public:
			virtual unsigned int OnExecute(PThread Thread) = 0;
	};

	typedef TVThreadFunc * PVThreadFunc;

	class TThread
	{
		public:
			TThread();
			virtual ~TThread();

			DWORD ID;

			int ThreadIndex;

			
#ifdef __GNUC__
			pthread_t t;
#else
			HANDLE ThreadHandle;
#endif

			PVThreadFunc Func;


			int Terminated;

			void SetTerminated(int _Terminated);
			int GetTerminated();

			void SetThreadIndex(int _ThreadIndex);
			size_t GetThreadIndex();
		
			void Start();
			void Final();

			virtual unsigned int OnExecute();

			void SetFunc(PVThreadFunc _Func);

			static void NanoSleep(unsigned int sec, int nanosec);

	};

	
	class TEvent
	{
		public:
			int Inited;
#ifdef __GNUC__
			//TCriticalS cs;
			pthread_cond_t cnd;
#else
			HANDLE e;
#endif

			TEvent();
			virtual ~TEvent();

			int Init();
			void Fin();

			void Run();

#ifdef __GNUC__
			void Wait(TCriticalS & cs);
#else
			void Wait();
#endif
	
	};

	PCriticalS _G_TCriticalS = GetGCriticalSection();

	TSingleton<TOnThreadCreateEvent> GOnThreadCreateEvent;
	
	void AddOnThreadCreateEvent(POnThreadCreateEvent Event)
	{
		GOnThreadCreateEvent.Get()->AddEvent(Event);
	}

	TThread::TThread()
	{
		Func = NULL;

		Terminated = 0;

		ThreadIndex = -1;
	}

	TThread::~TThread()
	{
		
	}

	void TOnThreadCreateEvent::AddEvent(POnThreadCreateEvent Event)
	{
		OnThreadCreate.push_back(Event);
	}

	TTLSint _G_CurrentThreadIdx;
	int GetCurrentThreadIdx()
	{
		return _G_CurrentThreadIdx.Get();
	}

	TOnThreadCreateEvent::TOnThreadCreateEvent()
	{
		ThreadCreated.resize(1);
		ThreadCreated[0] = 0;
	}
	
	TOnThreadCreate::TOnThreadCreate()
	{

		TCriticalSLocker Locker(GetGCriticalSection());

		size_t i, sz, ti;
		vector<int> & larray = GOnThreadCreateEvent.Get()->ThreadCreated;
		ti = 0;
		sz = larray.size();
		for(i = 0 ; i < sz ; i++)
		{
			if(larray[i] != (int)i)
			{
				ti = i;
				break;
			}
		}

		if(larray[ti] == -1)
		{
			_G_CurrentThreadIdx.Set((int)ti);
			larray[ti] = (int)ti;
		}
		else
		{
			larray.resize(sz + 1);
			larray[sz] = (int)sz;
			_G_CurrentThreadIdx.Set((int)sz);
		}
		
	}

	TOnThreadCreate::~TOnThreadCreate()
	{
		vector<int> & larray = GOnThreadCreateEvent.Get()->ThreadCreated;
		larray[GetCurrentThreadIdx()] = -1;
		_G_CurrentThreadIdx.Set(0);
	}




#ifdef __GNUC__
	void * TThread_ThreadFunc(void * val)
	{
		unsigned int ret = 0;
		PThread th = (PThread)val;

		TOnThreadCreate OnThreadCreate; 


		if(th)
		{
			ret = th->OnExecute();
			if(th->Func) ret = th->Func->OnExecute(th);
		}

		return (void *)ret;
	}
#else
	DWORD WINAPI TThread_ThreadFunc(void * val)
	{
		unsigned int ret;
		PThread th = (PThread)val;

		TOnThreadCreate OnThreadCreate; 


		if(th)
		{
			th->SetThreadIndex(GetCurrentThreadIdx());
			ret = th->OnExecute();
			if(th->Func) ret = th->Func->OnExecute(th);
		}

		return ret;
	}
#endif

	void TThread::SetTerminated(int _Terminated)
	{
		Terminated = _Terminated;
	}

	int TThread::GetTerminated()
	{
		return Terminated;
	}

	void TThread::SetThreadIndex(int _ThreadIndex)
	{
		ThreadIndex = _ThreadIndex;
	}

	size_t TThread::GetThreadIndex()
	{
		return ThreadIndex;
	}

	void TThread::Start()
	{
#ifdef __GNUC__
		pthread_create(&t,NULL,TThread_ThreadFunc,this);
#else
		ThreadHandle = CreateThread(NULL,0, &TThread_ThreadFunc,this, 0, &ID);
		//SetThreadPriority(
		//	  ThreadHandle,
		//	  THREAD_PRIORITY_ABOVE_NORMAL
		//	);

#endif
	}


	void TThread::Final()
	{
		Terminated = 1;

#ifdef __GNUC__
		pthread_join(t,NULL);
#else
		WaitForSingleObject(ThreadHandle, INFINITE);
		CloseHandle(ThreadHandle);
#endif
	}

	unsigned int TThread::OnExecute()
	{
		

		return 0;
	}

	void TThread::SetFunc(PVThreadFunc _Func)
	{
		Func = _Func;
	}

	void TThread::NanoSleep(unsigned int sec, int nanosec)
	{
#ifdef __GNUC__
		timespec ts;
		ts.tv_sec = sec;
		ts.tv_nsec = nanosec;
		nanosleep(&ts, NULL);
#else
		__int64 tm;

		tm = sec * 1000 + nanosec / (1000 * 1000);

		SleepEx((unsigned int)tm, TRUE);
		//Sleep((unsigned int)tm);
#endif
	}


	TEvent::TEvent()
	{
		Inited = 0;
	}

	TEvent::~TEvent()
	{
		Fin();
	}

	int TEvent::Init()
	{
		Fin();

#ifdef __GNUC__
		int ret;
		ret = pthread_cond_init(&cnd, NULL);
		if(ret != 0)
		{
			return 0;
		}
#else
		e = CreateEvent(NULL, 0, TRUE, NULL);
		ResetEvent(e);

#endif


		Inited = 1;
		return 1;
	}

	void TEvent::Fin()
	{
		if(Inited)
		{
#ifdef __GNUC__
			pthread_cond_destroy(&cnd);
#else
			CloseHandle(e);
#endif
		
			Inited = 0;
		}
	}

	void TEvent::Run()
	{
#ifdef __GNUC__
		int ret;
		//cs.Lock();
		
		ret = pthread_cond_signal(&cnd);

		//cs.UnLock();
#else
		unsigned int ret;
		ret = SetEvent(e);

		static int a;
		a++;
#endif
	}


#ifdef __GNUC__
	void TEvent::Wait(TCriticalS & cs)
	{
		int ret;
		//cs.Lock();
		
		ret = pthread_cond_wait(&cnd, &cs.GetMutex());

		//cs.UnLock();
	}

#else

	void TEvent::Wait()
	{
		unsigned int ret;
		ret = WaitForSingleObject(e, INFINITE);

		if(WAIT_ABANDONED == ret)
		{
			static int a;
			a++;
		}
		if(WAIT_OBJECT_0 == ret)
		{
			static int a;
			a++;
		}
		if(WAIT_TIMEOUT == ret)
		{
			static int a;
			a++;
		}
		if(WAIT_FAILED == ret)
		{
			static int a;
			a++;
		}
	}

#endif

//
	template<class T> class TList
	{
		public:
			struct _Item;
			typedef _Item * Item;
			typedef const _Item * cItem;

			struct _Item
			{
				T Data;
				Item Next;
				Item Prev;

				_Item()
				{
					Next = this;
					Prev = this;
				}

				inline void Remove()
				{
					Next->Prev = Prev;
					Prev->Next = Next;
					Next = this;
					Prev = this;
				}

				~_Item()
				{
					Remove();
				}

				inline void InsertNext(Item nxt)
				{
					nxt->Next->Prev = nxt->Prev;
					nxt->Prev->Next = nxt->Next;

					nxt->Prev = this;
					nxt->Next = Next;
					Next->Prev = nxt;
					Next = nxt;
				}

				inline void Append(Item nxt)
				{
					nxt->Prev->Next = Next;
					nxt->Prev = this;

					Next->Prev = nxt->Prev;
					Next = nxt;
				}

			};

			class TAList
			{
				public:
					Item Root;

					TAList()
					{
						Root = new _Item;
					}

					~TAList()
					{
						while(Root != Root->Next)
						{
							delete Root->Next;
						}

						delete Root;
					}

					inline int isEmpty()
					{
						return Root == Root->Next;
					}

					inline Item RemoveTop()
					{
						if(isEmpty()) return NULL;
						else
						{
							Item tmp;
							tmp = Root->Next;
							Root->Next->Remove();
							return tmp;
						}
					}
			};

			TAList LocalPool;
			TAList * LocalPool2;
			int UseLocalPool;

			Item Root;

			typedef vector<TAList *> TAListVec;

			class TOIF
			{
				public:
			
					TAListVec ObjPools;

					TOIF()
					{
						ObjPools.resize(1);
						ObjPools[0] = new TAList;
					}

					~TOIF()
					{
						size_t i, sz;
						sz = ObjPools.size();
						for(i = 0 ; i < sz ; i++)
						{
							delete ObjPools[i];
						}
					}
			};

			static inline TAListVec & GetObjPools()
			{
				static TOIF OPIF;
				return OPIF.ObjPools;
			}

			static inline void OnThreadCreateEvent(int idx)
			{
				TAListVec & ObjPools = GetObjPools();
				//Long idx = GetCurrentThreadIdx();

				if(idx >= (int)ObjPools.size())
				{
					ObjPools.resize(idx + 1);
					ObjPools[idx] = new TAList;
				}
			}

			inline TAList * GetObjPool()
			{
				//Long idx = GetCurrentThreadIdx();
				//return ObjPools[idx];

				TAList * lst;

				if(LocalPool2)
					lst = LocalPool2;
				else if(UseLocalPool)
					lst = &LocalPool;
				else
				{
					TAListVec & ObjPools = GetObjPools();
					lst = ObjPools[0];
				}

				return lst;
			}

			inline int isEmpty()
			{
				if(Root->Next == Root) return 1;
				return 0;
			}

			inline Item GetNewListItem()
			{
				TAList * ObjPool = GetObjPool();
				
				if(ObjPool->isEmpty())
				{
					Item p = new _Item;

					/*
					if((TSize)(void *)p == 0x01346D48)
					{
						static int a;
						a++;
					}
					*/

					return p;
				}
				else
				{
					Item p = ObjPool->RemoveTop();

					/*
					if((TSize)(void *)p == 0x01346D48)
					{
						static int a;
						a++;
					}
					*/

					return p;
				}

			}

			inline Item RemoveTop()
			{
				if(isEmpty()) return NULL;
				else
				{
					Item tmp;
					tmp = Root->Next;
					//Root->Next->Remove();
					Remove(Root->Next);
					return tmp;
				}
			}

			T pop_front()
			{
				return RemoveTop()->Data;
			}

			inline Item RemoveTail()
			{
				if(isEmpty()) return NULL;
				else
				{
					Item tmp;
					tmp = Root->Prev;
					//Root->Prev->Remove();
					Remove(Root->Prev);
					return tmp;
				}
			}

			T pop_back()
			{
				return RemoveTail()->Data;
			}

			inline Item InsertTop(const T & Data)
			{
				Item tmp = GetNewListItem();
				tmp->Data = Data;

				Root->InsertNext(tmp);
				return Root->Next;
			}

			inline Item push_front(const T & Data)
			{
				return InsertTop(Data);
			}

			inline Item InsertTail(const T & Data)
			{
				Item tmp = GetNewListItem();
				tmp->Data = Data;

				Root->Prev->InsertNext(tmp);
				return Root->Prev;
			}

			inline Item push_back(const T & Data)
			{
				Item p = InsertTail(Data);

				return p;
			}

			inline Item Insert(Item itm, const T & Data)
			{
				Item tmp = GetNewListItem();
				tmp->Data = Data;

				itm->Prev->InsertNext(tmp);
				return itm;
			}

			inline Item TopItem()
			{
				return Root->Next;
			}

			inline Item TailItem()
			{
				return Root->Prev;
			}

			inline Item RootItem()
			{
				return Root;
			}

			class TOnInitEnd
			{
				public:
					TOnInitEnd()
					{
						//GetSystem();
						AddOnThreadCreateEvent(OnThreadCreateEvent);
						GetObjPools();

					
					}

					~TOnInitEnd()
					{
						
					}
			};

			inline void Remove(Item tmp)
			{
				TAList * lst;
				lst = GetObjPool();
				lst->Root->InsertNext(tmp);
				
			}

			inline Item begin()
			{
				return Root->Next;
			}

			inline Item end()
			{
				return Root;
			}

			inline cItem begin() const
			{
				return Root->Next;
			}

			inline cItem end() const
			{
				return Root;
			}


			TList()
			{
				static TOnInitEnd OnInitEnd;

				LocalPool2 = NULL;
				UseLocalPool = 1;
				Root = GetNewListItem();
				
			}

			TList(const TList & lst)
			{
				static TOnInitEnd OnInitEnd;

				LocalPool2 = NULL;
				UseLocalPool = True;
				Root = GetNewListItem();

				CopyFrom(lst);
			}

			~TList()
			{
				if(Root)
				{
					TAList * lst = GetObjPool();
					lst->Root->Append(Root);
					Root = NULL;
				}
			}

			void CopyTo(TList & lst)
			{
				Item tmp, Root;
				Root = end();
				tmp = begin();
				while(tmp != Root)
				{
					lst.push_back(tmp->Data);

					tmp = tmp->Next;
				}
			}

			void CopyFrom(const TList & lst)
			{
				cItem tmp, Root;
				Root = lst.end();
				tmp = lst.begin();
				while(tmp != Root)
				{
					push_back(tmp->Data);

					tmp = tmp->Next;
				}
			}

			void SetUseLocalPool(int _UseLocalPool)
			{
				UseLocalPool = _UseLocalPool;
			}

			int GetUseLocalPool()
			{
				return UseLocalPool;
			}

			void SetLocalPool(TAList * alist)
			{
				LocalPool2 = alist;
			}

			TAList * GetLocalPool()
			{
				return LocalPool2;
			}

			inline size_t size()
			{
				size_t sz = 0;
				Item Root, tmp;

				Root = RootItem();
				tmp = TopItem();

				while(tmp != Root)
				{
					sz++;
					tmp = tmp->Next;
				}

				return sz;
			}

			void AllRemove()
			{
				TAList * lst = GetObjPool();

				lst->Root->Append(Root);

				Root = GetNewListItem();
			}

			void clear()
			{
				AllRemove();
			}

			void AllDestroy()
			{
				Item Root, tmp;

				Root = end();
				tmp = begin();

				while(tmp != Root)
				{
					tmp->Data.~T();

					tmp = tmp->Next;
				}
			}

			void clearWithDestroy()
			{
				AllDestroy();
				AllRemove();
			}

			TList & operator=(const TList & lst)
			{
				CopyTo(*this);
				return *this;
			}

			//Item begin()

	
	};

//
	class TJob;
	typedef TJob * PJob;

	class TJob
	{
		public:

			int ThreadIndex;

			TJob();
			virtual ~TJob();
			
			void SetThreadIndex(int _ThreadIndex);
			int GetThreadIndex();

			virtual unsigned int OnExecute();
			virtual unsigned int OnEnd();
	};

	class TWorkerManager;
	typedef TWorkerManager * PWorkerManager;

	class TWorkerThread;
	typedef TWorkerThread * PWorkerThread;

	class TWorkerThread : public TThread
	{
		public:
			PWorkerManager Manager;
			TCriticalS ThreadCS;
			TEvent Evt;
			PJob Job;
			int Running;
			int Locked;
			int LightMode;
			int Waited;

			TWorkerThread();
			virtual ~TWorkerThread();

			TCriticalS & GetThreadCS();
			TEvent & GetEvt();
			void SetManager(PWorkerManager _Manager);
			int GetRunning();
			void SetLocked(int _Locked);
			int GetLocked();
			void SetLightMode(int _LightMode);
			int GetLightMode();
			void SetWaited(int _Waited);
			int GetWaited();

			virtual unsigned int OnExecute();
	};

	class TWorkerManager
	{
		public:
			size_t ThreadSize;
			size_t WaitCount;
			TCriticalS WaitCS;
			TEvent Evt;

			TCriticalS QueueCS;
			TList<PJob> JobQueue;

			TCriticalS EndQueueCS;
			TList<PJob> EndJobQueue;

			vector<PWorkerThread> WThread;

			int Locked;
			int LightMode;

			TWorkerManager();
			virtual ~TWorkerManager();

			int Init(size_t _ThreadSize, int _LightMode);
			void Fin();

			vector<PWorkerThread> & GetWThread();

			int AddJobToQueue(PJob job);
			PJob GetJobFromQueue();
			int AddEndJob(PJob job);

			int Run();
			void Wait();

			void WaitRelease();

			virtual void Poll();

			int Idle();
	
	};


	TJob::TJob()
	{
	
	}

	TJob::~TJob()
	{
	
	}

	void TJob::SetThreadIndex(int _ThreadIndex)
	{
		ThreadIndex = _ThreadIndex;
	}
	
	int TJob::GetThreadIndex()
	{
		return ThreadIndex;
	}

	unsigned int TJob::OnExecute()
	{
		return 0;
	}

	unsigned int TJob::OnEnd()
	{
		return 0;
	}


	
	TWorkerThread::TWorkerThread()
	{
		Job = NULL;
		Running = 0;
		Locked = 0;
		LightMode = 0;
		Waited = 0;
	}

	TWorkerThread::~TWorkerThread()
	{
	
	}

	TCriticalS & TWorkerThread::GetThreadCS()
	{
		return ThreadCS;
	}

	TEvent & TWorkerThread::GetEvt()
	{
		return Evt;
	}

	void TWorkerThread::SetManager(PWorkerManager _Manager)
	{
		Manager = _Manager;
	}

	int TWorkerThread::GetRunning()
	{
		return Running;
	}

	void TWorkerThread::SetLocked(int _Locked)
	{
		Locked = _Locked;
	}

	int TWorkerThread::GetLocked()
	{
		return Locked;
	}

	void TWorkerThread::SetLightMode(int _LightMode)
	{
		LightMode = _LightMode;
	}

	int TWorkerThread::GetLightMode()
	{
		return LightMode;
	}

	void TWorkerThread::SetWaited(int _Waited)
	{
		Waited = _Waited;
	}

	int TWorkerThread::GetWaited()
	{
		return Waited;
	}

	unsigned int TWorkerThread::OnExecute()
	{
		unsigned int NULLCount = 0;

		if(LightMode == 0)
		{
#ifdef __GNUC__
			GetThreadCS().Lock();
			GetEvt().Wait(GetThreadCS());
#else
			GetEvt().Wait();
#endif
		}

		while(1)
		{
			//GetThreadCS().Lock();
			Running = 1;
			
			Job = Manager->GetJobFromQueue();
			while(Job)
			{
				NULLCount = 0;

				Job->SetThreadIndex((int)GetThreadIndex());
				Job->OnExecute();
				Manager->AddEndJob(Job);

				Job = Manager->GetJobFromQueue();
			}

			NULLCount++;

			

			Running = 0;
			//GetThreadCS().UnLock();

			if(LightMode == 0)
			{
				Manager->WaitRelease();
#ifdef __GNUC__
				GetThreadCS().Lock();
				GetEvt().Wait(GetThreadCS());
#else
				GetEvt().Wait();
#endif
			}
			
			if(LightMode && NULLCount > 10)
			{
				TThread::NanoSleep(0, 1000 * 1000);

				if(GetWaited())
				{
					Manager->WaitRelease();
				}
			}

			

			//TThread::NanoSleep(0, 0);

			if(GetTerminated())
			{
				break;
			}
		}

		return 0;
	}


	TWorkerManager::TWorkerManager()
	{
		Locked = 0;
		LightMode = 0;
		ThreadSize = 0;
		WaitCount = 0;
	}

	TWorkerManager::~TWorkerManager()
	{
		Fin();
	}

	int TWorkerManager::Init(size_t _ThreadSize, int _LightMode)
	{
		Fin();

		ThreadSize = _ThreadSize;
		LightMode = _LightMode;
		WThread.resize(_ThreadSize);

		Evt.Init();

		size_t i, isz;
		isz = _ThreadSize;

		for(i = 0 ; i< isz ; i++)
		{
			WThread[i] = new TWorkerThread;
			WThread[i]->SetManager(this);

			WThread[i]->GetEvt().Init();

			//WThread[i]->GetThreadCS().Lock();

			WThread[i]->Start();

			WThread[i]->SetLocked(1);
			WThread[i]->SetLightMode(_LightMode);
		}

		WaitCount = 0;
		Locked = 1;
	
		return 1;
	}

	void TWorkerManager::Fin()
	{
		size_t i, isz;

		if(Locked == 0)
		{
			Wait();
		}

		Locked = 0;

		isz = WThread.size();
		for(i = 0 ; i < isz ; i++)
		{
			WThread[i]->SetLocked(0);
			//WThread[i]->GetThreadCS().UnLock();

			WThread[i]->SetTerminated(1);
			WThread[i]->GetEvt().Run();
			WThread[i]->Final();
			
		}

		

		for(i = 0 ; i < isz ; i++)
		{
			WThread[i]->GetEvt().Fin();
			delete WThread[i];
			WThread[i] = NULL;
		}

		Evt.Fin();

		WThread.clear();
	}

	vector<PWorkerThread> & TWorkerManager::GetWThread()
	{
		return WThread;
	}

	int TWorkerManager::AddJobToQueue(PJob job)
	{
		TCriticalSLocker Locker(&QueueCS);

		JobQueue.push_back(job);
	
		return 1;
	}

	PJob TWorkerManager::GetJobFromQueue()
	{
		TCriticalSLocker Locker(&QueueCS);

		if(JobQueue.isEmpty())
		{
			return NULL;
		}

		PJob job = JobQueue.begin()->Data;
		JobQueue.pop_front();
		return job;
	}

	int TWorkerManager::AddEndJob(PJob job)
	{
		TCriticalSLocker Locker(&EndQueueCS);

		EndJobQueue.push_back(job);

		return 1;
	}

	int TWorkerManager::Run()
	{
		if(Locked == 0)
		{
			return 1;
		}

		Locked = 0;

		WaitCount = ThreadSize;

		size_t i, isz;
		isz = WThread.size();

		for(i = 0 ; i < isz ; i++)
		{
			WThread[i]->SetLocked(0);
			WThread[i]->SetWaited(0);
			//WThread[i]->GetThreadCS().UnLock();
			WThread[i]->GetEvt().Run();

		}

		//for(i = 0 ; i < isz ; i++)
		//{
		//	while(WThread[i]->GetRunning() == 0)
		//	{
		//		TThread::NanoSleep(0, 0);
		//	}
		//}

		//Sleep(0);
	
		return 1;
	}

	void TWorkerManager::Wait()
	{
		if(Locked == 1)
		{
			return;
		}

		size_t i, isz;
		isz = WThread.size();

		Locked = 1;

		for(i = 0 ; i < isz ; i++)
		{
			WThread[i]->SetWaited(1);
		}

		//WaitCount = ThreadSize;

#ifdef __GNUC__
		{
			//TCriticalSLocker Locker(&WaitCS);

			WaitCS.Lock();

			Evt.Wait(WaitCS);
		}
#else
		Evt.Wait();
#endif


		/*
		for(i = 0 ; i < isz ; i++)
		{
			if(WThread[i]->GetLocked() == 0)
			{
				WThread[i]->SetLocked(1);
				//WThread[i]->GetThreadCS().Lock();
			}
		}
		
		for(i = 0 ; i < isz ; i++)
		{
			while(WThread[i]->GetRunning() == 1)
			{
				TThread::NanoSleep(0, 0);
			}
		}
		//*/
	}

	void TWorkerManager::WaitRelease()
	{
		TCriticalSLocker Locker(&WaitCS);

		if(WaitCount > 0)
		{
			WaitCount--;
			if(WaitCount <= 0)
			{
				Evt.Run();
			}
		}
	}

	void TWorkerManager::Poll()
	{
		TCriticalSLocker Locker(&EndQueueCS);

		PJob job;
		while(EndJobQueue.isEmpty() == 0)
		{
			job = EndJobQueue.begin()->Data;
			EndJobQueue.pop_front();

			job->OnEnd();
		}
	}

	int TWorkerManager::Idle()
	{
		TCriticalSLocker Locker(&QueueCS);
		TCriticalSLocker Locker2(&EndQueueCS);

		if(JobQueue.size() <= 0 && EndJobQueue.size() <= 0)
		{
			return 1;
		}

		return 0;
	}


//
#define RANDOM_SIZE 8

#include <sys/types.h>
#include <sys/timeb.h>
static struct _timeb start, end;
static void reset_and_start_timer()
{
	_ftime(&start);
}
static double get_elapsed_seconds()
{
	_ftime(&end);
	return (1000 * (end.time - start.time) + (end.millitm - start.millitm)) / 1000.0;
}

static long long drand48_x[RANDOM_SIZE] = {
		0x1234ABCD330E, 0x1234ABCD331E, 0x1234ABCD332E, 0x1234ABCD333E,
		0x1234ABCD334E, 0x1234ABCD335E, 0x1234ABCD336E, 0x1234ABCD337E
	};
static double drand48(int i)
{
	drand48_x[i] = drand48_x[i] * 0x5DEECE66D + 0xB;
	return (drand48_x[i] & 0xFFFFFFFFFFFF) * (1.0 / 281474976710656.0);
}
#define M_PI 3.14159265358979

#define WIDTH        256
#define HEIGHT       256
#define NSUBSAMPLES  2
#define NAO_SAMPLES  8

typedef struct _vec
{
	double x;
	double y;
	double z;
} vec;


typedef struct _Isect
{
	double t;
	vec    p;
	vec    n;
	int    hit; 
} Isect;

typedef struct _Sphere
{
	vec    center;
	double radius;

} Sphere;

typedef struct _Plane
{
	vec    p;
	vec    n;

} Plane;

typedef struct _Ray
{
	vec    org;
	vec    dir;
} Ray;

Sphere spheres[3];
Plane  plane;

static double vdot(vec v0, vec v1)
{
	return v0.x * v1.x + v0.y * v1.y + v0.z * v1.z;
}

static void vcross(vec *c, vec v0, vec v1)
{
	c->x = v0.y * v1.z - v0.z * v1.y;
	c->y = v0.z * v1.x - v0.x * v1.z;
	c->z = v0.x * v1.y - v0.y * v1.x;
}

static void vnormalize(vec *c)
{
	double length = sqrt(vdot((*c), (*c)));

	if (length > 1.0e-17) {
		c->x /= length;
		c->y /= length;
		c->z /= length;
	}
}

void
ray_sphere_intersect(Isect *isect, const Ray *ray, const Sphere *sphere)
{
	vec rs;
	double B, C, D;

	rs.x = ray->org.x - sphere->center.x;
	rs.y = ray->org.y - sphere->center.y;
	rs.z = ray->org.z - sphere->center.z;

	B = vdot(rs, ray->dir);
	C = vdot(rs, rs) - sphere->radius * sphere->radius;
	D = B * B - C;

	if (D > 0.0) {
		double t = -B - sqrt(D);

		if ((t > 0.0) && (t < isect->t)) {
			isect->t = t;
			isect->hit = 1;

			isect->p.x = ray->org.x + ray->dir.x * t;
			isect->p.y = ray->org.y + ray->dir.y * t;
			isect->p.z = ray->org.z + ray->dir.z * t;

			isect->n.x = isect->p.x - sphere->center.x;
			isect->n.y = isect->p.y - sphere->center.y;
			isect->n.z = isect->p.z - sphere->center.z;

			vnormalize(&(isect->n));
		}
	}
}

void
ray_plane_intersect(Isect *isect, const Ray *ray, const Plane *plane)
{
	double d, v, t;
	d = -vdot(plane->p, plane->n);
	v = vdot(ray->dir, plane->n);

	if (fabs(v) < 1.0e-17) return;

	t = -(vdot(ray->org, plane->n) + d) / v;

	if ((t > 0.0) && (t < isect->t)) {
		isect->t = t;
		isect->hit = 1;

		isect->p.x = ray->org.x + ray->dir.x * t;
		isect->p.y = ray->org.y + ray->dir.y * t;
		isect->p.z = ray->org.z + ray->dir.z * t;

		isect->n = plane->n;
	}
}

void
orthoBasis(vec *basis, vec n)
{
	basis[2] = n;
	basis[1].x = 0.0; basis[1].y = 0.0; basis[1].z = 0.0;

	if ((n.x < 0.6) && (n.x > -0.6)) {
		basis[1].x = 1.0;
	} else if ((n.y < 0.6) && (n.y > -0.6)) {
		basis[1].y = 1.0;
	} else if ((n.z < 0.6) && (n.z > -0.6)) {
		basis[1].z = 1.0;
	} else {
		basis[1].x = 1.0;
	}

	vcross(&basis[0], basis[1], basis[2]);
	vnormalize(&basis[0]);

	vcross(&basis[1], basis[2], basis[0]);
	vnormalize(&basis[1]);
}

void
ambient_occlusion(vec *col, const Isect *isect, int threadIdx)
{
	int    i, j;
	int    ntheta = NAO_SAMPLES;
	int    nphi   = NAO_SAMPLES;
	double eps = 0.0001;

	vec p;
	vec basis[3];
	double occlusion = 0.0;

	p.x = isect->p.x + eps * isect->n.x;
	p.y = isect->p.y + eps * isect->n.y;
	p.z = isect->p.z + eps * isect->n.z;

	orthoBasis(basis, isect->n);

	for (j = 0; j < ntheta; j++) {
		for (i = 0; i < nphi; i++) {
			double theta = sqrt(drand48(threadIdx));
			double phi   = 2.0 * M_PI * drand48(threadIdx);

			double x = cos(phi) * theta;
			double y = sin(phi) * theta;
			double z = sqrt(1.0 - theta * theta);

			// local -> global
			double rx = x * basis[0].x + y * basis[1].x + z * basis[2].x;
			double ry = x * basis[0].y + y * basis[1].y + z * basis[2].y;
			double rz = x * basis[0].z + y * basis[1].z + z * basis[2].z;

			Ray ray;
			Isect occIsect;

			ray.org = p;
			ray.dir.x = rx;
			ray.dir.y = ry;
			ray.dir.z = rz;

			occIsect.t   = 1.0e+17;
			occIsect.hit = 0;

			ray_sphere_intersect(&occIsect, &ray, &spheres[0]); 
			ray_sphere_intersect(&occIsect, &ray, &spheres[1]); 
			ray_sphere_intersect(&occIsect, &ray, &spheres[2]); 
			ray_plane_intersect (&occIsect, &ray, &plane); 

			if (occIsect.hit) occlusion += 1.0;

		}
	}

	occlusion = (ntheta * nphi - occlusion) / (double)(ntheta * nphi);

	col->x = occlusion;
	col->y = occlusion;
	col->z = occlusion;
}

unsigned char
clamp(double f)
{
	int i = (int)(f * 255.5);

	if (i < 0) i = 0;
	if (i > 255) i = 255;

	return (unsigned char)i;
}

//
void renderLine(unsigned char *img, int y, double * fimg, int w, int h, int nsubsamples, int threadIdx)
{
	int x;
	int u, v;

	for (x = 0; x < w; x++) {
		for (v = 0; v < nsubsamples; v++) {
			for (u = 0; u < nsubsamples; u++) {
				double px = (x + (u / (double)nsubsamples) - (w / 2.0)) / (w / 2.0);
				double py = -(y + (v / (double)nsubsamples) - (h / 2.0)) / (h / 2.0);

				Ray ray;
				Isect isect;

				ray.org.x = 0.0;
				ray.org.y = 0.0;
				ray.org.z = 0.0;

				ray.dir.x = px;
				ray.dir.y = py;
				ray.dir.z = -1.0;
				vnormalize(&(ray.dir));

				isect.t   = 1.0e+17;
				isect.hit = 0;

				ray_sphere_intersect(&isect, &ray, &spheres[0]);
				ray_sphere_intersect(&isect, &ray, &spheres[1]);
				ray_sphere_intersect(&isect, &ray, &spheres[2]);
				ray_plane_intersect (&isect, &ray, &plane);

				if (isect.hit) {
					vec col;
					ambient_occlusion(&col, &isect, threadIdx);

					fimg[3 * (y * w + x) + 0] += col.x;
					fimg[3 * (y * w + x) + 1] += col.y;
					fimg[3 * (y * w + x) + 2] += col.z;
				}

			}
		}

		fimg[3 * (y * w + x) + 0] /= (double)(nsubsamples * nsubsamples);
		fimg[3 * (y * w + x) + 1] /= (double)(nsubsamples * nsubsamples);
		fimg[3 * (y * w + x) + 2] /= (double)(nsubsamples * nsubsamples);

		img[3 * (y * w + x) + 0] = clamp(fimg[3 *(y * w + x) + 0]);
		img[3 * (y * w + x) + 1] = clamp(fimg[3 *(y * w + x) + 1]);
		img[3 * (y * w + x) + 2] = clamp(fimg[3 *(y * w + x) + 2]);
	}
}

	class TRenderLineJob : public TJob
	{
		public:


			unsigned char *img;
			int y;
			double * fimg;
			int w;
			int h;
			int nsubsamples;

			void SetParam(unsigned char *img,
						int y, double * fimg, int w, int h,
						int nsubsamples);

			virtual unsigned int OnExecute();

			virtual unsigned int OnEnd();
								   
	};

	void TRenderLineJob::SetParam(unsigned char *img,
						int y, double * fimg, int w, int h,
						int nsubsamples)
	{
		this->img = img;
		this->y = y;
		this->fimg = fimg;
		this->w = w;
		this->h = h;
		this->nsubsamples = nsubsamples;
	}

	unsigned int TRenderLineJob::OnExecute()
	{
		renderLine(img, y, fimg, w, h, nsubsamples, GetThreadIndex());

		return 0;
	}

	unsigned int TRenderLineJob::OnEnd()
	{
	
		return 0;
	}

//
TWorkerManager Worker;
TRenderLineJob Jobs[HEIGHT];


//
void
render(unsigned char *img, int w, int h, int nsubsamples)
{
	int y;

	double *fimg = (double *)malloc(sizeof(double) * w * h * 3);
	memset((void *)fimg, 0, sizeof(double) * w * h * 3);

	/*
	for (y = 0; y < h; y++) {
		renderLine(img, y, fimg, w, h, nsubsamples, 0);
	}
	/*/
	for (y = 0; y < h; y++) {
		Jobs[y].SetParam(img, y, fimg, w, h, nsubsamples);
		Worker.AddJobToQueue(&Jobs[y]);
	}

	Worker.Run();
	Worker.Wait();

	Worker.Poll();

	//*/

	free(fimg);
}

//
/*
void
render(unsigned char *img, int w, int h, int nsubsamples)
{
	int x, y;
	int u, v;

	double *fimg = (double *)malloc(sizeof(double) * w * h * 3);
	memset((void *)fimg, 0, sizeof(double) * w * h * 3);

	for (y = 0; y < h; y++) {
		for (x = 0; x < w; x++) {

			for (v = 0; v < nsubsamples; v++) {
				for (u = 0; u < nsubsamples; u++) {
					double px = (x + (u / (double)nsubsamples) - (w / 2.0)) / (w / 2.0);
					double py = -(y + (v / (double)nsubsamples) - (h / 2.0)) / (h / 2.0);

					Ray ray;
					Isect isect;

					ray.org.x = 0.0;
					ray.org.y = 0.0;
					ray.org.z = 0.0;

					ray.dir.x = px;
					ray.dir.y = py;
					ray.dir.z = -1.0;
					vnormalize(&(ray.dir));

					isect.t   = 1.0e+17;
					isect.hit = 0;

					ray_sphere_intersect(&isect, &ray, &spheres[0]);
					ray_sphere_intersect(&isect, &ray, &spheres[1]);
					ray_sphere_intersect(&isect, &ray, &spheres[2]);
					ray_plane_intersect (&isect, &ray, &plane);

					if (isect.hit) {
						vec col;
						ambient_occlusion(&col, &isect);

						fimg[3 * (y * w + x) + 0] += col.x;
						fimg[3 * (y * w + x) + 1] += col.y;
						fimg[3 * (y * w + x) + 2] += col.z;
					}

				}
			}

			fimg[3 * (y * w + x) + 0] /= (double)(nsubsamples * nsubsamples);
			fimg[3 * (y * w + x) + 1] /= (double)(nsubsamples * nsubsamples);
			fimg[3 * (y * w + x) + 2] /= (double)(nsubsamples * nsubsamples);

			img[3 * (y * w + x) + 0] = clamp(fimg[3 *(y * w + x) + 0]);
			img[3 * (y * w + x) + 1] = clamp(fimg[3 *(y * w + x) + 1]);
			img[3 * (y * w + x) + 2] = clamp(fimg[3 *(y * w + x) + 2]);
		}
	}

	free(fimg);
}
*/

void
init_scene()
{
	spheres[0].center.x = -2.0;
	spheres[0].center.y =  0.0;
	spheres[0].center.z = -3.5;
	spheres[0].radius = 0.5;

	spheres[1].center.x = -0.5;
	spheres[1].center.y =  0.0;
	spheres[1].center.z = -3.0;
	spheres[1].radius = 0.5;

	spheres[2].center.x =  1.0;
	spheres[2].center.y =  0.0;
	spheres[2].center.z = -2.2;
	spheres[2].radius = 0.5;

	plane.p.x = 0.0;
	plane.p.y = -0.5;
	plane.p.z = 0.0;

	plane.n.x = 0.0;
	plane.n.y = 1.0;
	plane.n.z = 0.0;

}

void
saveppm(const char *fname, int w, int h, unsigned char *img)
{
	FILE *fp;

	fp = fopen(fname, "wb");
	assert(fp);

	fprintf(fp, "P6\n");
	fprintf(fp, "%d %d\n", w, h);
	fprintf(fp, "255\n");
	fwrite(img, w * h * 3, 1, fp);
	fclose(fp);
}

int
main(int argc, char **argv)
{
	unsigned char *img = (unsigned char *)malloc(WIDTH * HEIGHT * 3);
	int i, N = 20;
	double avg = 0.0;

	Worker.Init(4, 0);
	Worker.Run();
	Worker.Wait();

	for (i = 0; i < N; i++) {
		reset_and_start_timer();
		init_scene();
		render(img, WIDTH, HEIGHT, NSUBSAMPLES);
		avg += get_elapsed_seconds();
	}
	avg /= N;
	printf("Elapsed: %7.3f[sec] (N=%d)\n", avg, N);
	saveppm("ao.ppm", WIDTH, HEIGHT, img); 

	return 0;
}
